#include "runner.hpp"
#include <cstdlib>
#include <sstream>
#include <iostream>
#include <unistd.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <libssh/libssh.h>
#include <libssh/callbacks.h>
#include <termios.h>
#include <sys/select.h>
#include <algorithm>
#include <cctype>

namespace runner {

namespace {

// String trimming functions
void ltrim(std::string& s) {
    s.erase(s.begin(), std::find_if(s.begin(), s.end(), [](unsigned char ch) {
        return !std::isspace(ch);
    }));
}

void rtrim(std::string& s) {
    s.erase(std::find_if(s.rbegin(), s.rend(), [](unsigned char ch) {
        return !std::isspace(ch);
    }).base(), s.end());
}

void trim(std::string& s) {
    ltrim(s);
    rtrim(s);
}

// Safe version that handles nullptr
void trim(std::string* s) {
    if (s) {
        trim(*s);
    }
}

ssh_session ssh_connect_and_auth(const sSSHInfo* sshinfo, const std::map<std::string, std::string>& env, std::string* error) {
    ssh_session session = ssh_new();
    if (!session) {
        if (error) *error = "Failed to create SSH session.";
        return nullptr;
    }
    ssh_options_set(session, SSH_OPTIONS_HOST, sshinfo->host.c_str());
    if (!sshinfo->port.empty()) {
        int port = std::stoi(sshinfo->port);
        ssh_options_set(session, SSH_OPTIONS_PORT, &port);
    }
    if (!sshinfo->user.empty()) {
        ssh_options_set(session, SSH_OPTIONS_USER, sshinfo->user.c_str());
    }
    int rc = ssh_connect(session);
    if (rc != SSH_OK) {
        if (error) *error = std::string("SSH connection failed: ") + ssh_get_error(session);
        ssh_free(session);
        return nullptr;
    }
    rc = ssh_userauth_publickey_auto(session, nullptr, nullptr);
    if (rc != SSH_AUTH_SUCCESS) {
        auto it = env.find("SSHPASS");
        if (it != env.end()) {
            rc = ssh_userauth_password(session, nullptr, it->second.c_str());
        }
    }
    if (rc != SSH_AUTH_SUCCESS) {
        if (error) *error = std::string("SSH authentication failed: ") + ssh_get_error(session);
        ssh_disconnect(session);
        ssh_free(session);
        return nullptr;
    }
    return session;
}

} // namespace runner

// runner_local: non-interactive local command execution
bool runner_local::execute() {
    pid_t pid = fork();
    if (pid == -1) {
        std::cerr << "Failed to fork" << std::endl;
        mReturnCode = -1;
        return false;
    }
    if (pid == 0) {
        // Child process
        if (!mWorkingDir.empty()) {
            if (chdir(mWorkingDir.c_str()) != 0) {
                perror("chdir");
                exit(127);
            }
        }
        // Set environment variables
        for (const auto& [k, v] : mEnv) {
            setenv(k.c_str(), v.c_str(), 1);
        }
        // Build argv
        std::vector<char*> argv;
        argv.push_back(const_cast<char*>(mCommand.c_str()));
        for (auto& arg : mArgs) argv.push_back(const_cast<char*>(arg.c_str()));
        argv.push_back(nullptr);
        if (mSilent) {
            int fd = open("/dev/null", O_WRONLY);
            if (fd != -1) {
                dup2(fd, STDOUT_FILENO);
                dup2(fd, STDERR_FILENO);
                close(fd);
            }
        }

        std::cout << "Executing command: " << mCommand << std::endl;    
        std::cout << "Args: ";
            for (auto& arg : mArgs) std::cout << "[" << arg << "] ";
        std::cout << std::endl;

        execvp(mCommand.c_str(), argv.data());
        perror("execvp");
        exit(127);
    } else {
        int status = 0;
        waitpid(pid, &status, 0);
        if (WIFEXITED(status)) {
            mReturnCode = WEXITSTATUS(status);
        } else {
            mReturnCode = -1;
        }
        return mReturnCode == 0;
    }
}

// runner_local_interactive: interactive local command execution
bool runner_local_interactive::execute() {
    pid_t pid = fork();
    if (pid == -1) {
        std::cerr << "Failed to fork" << std::endl;
        mReturnCode = -1;
        return false;
    }
    if (pid == 0) {
        if (!mWorkingDir.empty()) {
            if (chdir(mWorkingDir.c_str()) != 0) {
                perror("chdir");
                exit(127);
            }
        }
        for (const auto& [k, v] : mEnv) {
            setenv(k.c_str(), v.c_str(), 1);
        }
        std::vector<char*> argv;
        argv.push_back(const_cast<char*>(mCommand.c_str()));
        for (auto& arg : mArgs) argv.push_back(const_cast<char*>(arg.c_str()));
        argv.push_back(nullptr);
        // Attach to terminal (no redirection)
        execvp(mCommand.c_str(), argv.data());
        perror("execvp");
        exit(127);
    } else {
        int status = 0;
        waitpid(pid, &status, 0);
        if (WIFEXITED(status)) {
            mReturnCode = WEXITSTATUS(status);
        } else {
            mReturnCode = -1;
        }
        return mReturnCode == 0;
    }
}

// runner_ssh: non-interactive SSH command execution
bool runner_ssh::execute() {
    std::string error;
    ssh_session session = ssh_connect_and_auth(&mSSHInfo, mEnv, &error);
    if (!session) {
        std::cerr << error << std::endl;
        mReturnCode = -1;
        return false;
    }
    ssh_channel channel = ssh_channel_new(session);
    if (!channel) {
        std::cerr << "Failed to create SSH channel." << std::endl;
        ssh_disconnect(session);
        ssh_free(session);
        mReturnCode = -1;
        return false;
    }
    if (ssh_channel_open_session(channel) != SSH_OK) {
        std::cerr << "Failed to open SSH channel: " << ssh_get_error(session) << std::endl;
        ssh_channel_free(channel);
        ssh_disconnect(session);
        ssh_free(session);
        mReturnCode = -1;
        return false;
    }
    std::stringstream cmd;
    cmd << mCommand;
    for (const auto& arg : mArgs) {
        cmd << " '" << arg << "'";
    }
    int rc = ssh_channel_request_exec(channel, cmd.str().c_str());
    if (rc != SSH_OK) {
        std::cerr << "SSH exec failed: " << ssh_get_error(session) << std::endl;
        ssh_channel_close(channel);
        ssh_channel_free(channel);
        ssh_disconnect(session);
        ssh_free(session);
        mReturnCode = -1;
        return false;
    }
    char buffer[256];
    int nbytes;
    while ((nbytes = ssh_channel_read(channel, buffer, sizeof(buffer), 0)) > 0) {
        if (!mSilent) std::cout.write(buffer, nbytes);
    }
    mReturnCode = ssh_channel_get_exit_status(channel);
    ssh_channel_send_eof(channel);
    ssh_channel_close(channel);
    ssh_channel_free(channel);
    ssh_disconnect(session);
    ssh_free(session);
    return mReturnCode == 0;
}

// runner_ssh_interactive: interactive SSH command execution
bool runner_ssh_interactive::execute() {
    std::string error;
    ssh_session session = ssh_connect_and_auth(&mSSHInfo, mEnv, &error);
    if (!session) {
        std::cerr << error << std::endl;
        mReturnCode = -1;
        return false;
    }
    ssh_channel channel = ssh_channel_new(session);
    if (!channel) {
        std::cerr << "Failed to create SSH channel." << std::endl;
        ssh_disconnect(session);
        ssh_free(session);
        mReturnCode = -1;
        return false;
    }
    if (ssh_channel_open_session(channel) != SSH_OK) {
        std::cerr << "Failed to open SSH channel: " << ssh_get_error(session) << std::endl;
        ssh_channel_free(channel);
        ssh_disconnect(session);
        ssh_free(session);
        mReturnCode = -1;
        return false;
    }
    std::stringstream cmd;
    cmd << mCommand;
    for (const auto& arg : mArgs) {
        cmd << " '" << arg << "'";
    }
    int rc = ssh_channel_request_pty(channel);
    if (rc != SSH_OK) {
        std::cerr << "Failed to request PTY: " << ssh_get_error(session) << std::endl;
        ssh_channel_close(channel);
        ssh_channel_free(channel);
        ssh_disconnect(session);
        ssh_free(session);
        mReturnCode = -1;
        return false;
    }
    rc = ssh_channel_request_shell(channel);
    if (rc != SSH_OK) {
        std::cerr << "Failed to request shell: " << ssh_get_error(session) << std::endl;
        ssh_channel_close(channel);
        ssh_channel_free(channel);
        ssh_disconnect(session);
        ssh_free(session);
        mReturnCode = -1;
        return false;
    }
    // Forward input/output between user and SSH channel
    fd_set fds;
    char buffer[256];
    while (true) {
        FD_ZERO(&fds);
        FD_SET(0, &fds); // stdin
        int ssh_fd = ssh_get_fd(session);
        FD_SET(ssh_fd, &fds);
        int maxfd = std::max(0, ssh_fd) + 1;
        int ret = select(maxfd, &fds, nullptr, nullptr, nullptr);
        if (ret < 0) break;
        if (FD_ISSET(0, &fds)) {
            int n = read(0, buffer, sizeof(buffer));
            if (n > 0) ssh_channel_write(channel, buffer, n);
        }
        if (FD_ISSET(ssh_fd, &fds)) {
            int n = ssh_channel_read(channel, buffer, sizeof(buffer), 0);
            if (n > 0) write(1, buffer, n);
            else if (n == 0) break;
        }
        if (ssh_channel_is_closed(channel)) break;
    }
    mReturnCode = ssh_channel_get_exit_status(channel);
    ssh_channel_send_eof(channel);
    ssh_channel_close(channel);
    ssh_channel_free(channel);
    ssh_disconnect(session);
    ssh_free(session);
    return mReturnCode == 0;
}

// runner_ssh_capture: SSH command execution with output capture
bool runner_ssh_capture::execute() {
    std::string error;
    ssh_session session = ssh_connect_and_auth(&mSSHInfo, mEnv, &error);
    if (!session) {
        std::cerr << error << std::endl;
        mReturnCode = -1;
        return false;
    }
    ssh_channel channel = ssh_channel_new(session);
    if (!channel) {
        std::cerr << "Failed to create SSH channel." << std::endl;
        ssh_disconnect(session);
        ssh_free(session);
        mReturnCode = -1;
        return false;
    }
    if (ssh_channel_open_session(channel) != SSH_OK) {
        std::cerr << "Failed to open SSH channel: " << ssh_get_error(session) << std::endl;
        ssh_channel_free(channel);
        ssh_disconnect(session);
        ssh_free(session);
        mReturnCode = -1;
        return false;
    }
    std::stringstream cmd;
    cmd << mCommand;
    for (const auto& arg : mArgs) {
        cmd << " '" << arg << "'";
    }
    int rc = ssh_channel_request_exec(channel, cmd.str().c_str());
    if (rc != SSH_OK) {
        std::cerr << "SSH exec failed: " << ssh_get_error(session) << std::endl;
        ssh_channel_close(channel);
        ssh_channel_free(channel);
        ssh_disconnect(session);
        ssh_free(session);
        mReturnCode = -1;
        return false;
    }
    char buffer[256];
    int nbytes;
    mOutput.clear();
    while ((nbytes = ssh_channel_read(channel, buffer, sizeof(buffer), 0)) > 0) {
        mOutput.append(buffer, nbytes);
    }
    mReturnCode = ssh_channel_get_exit_status(channel);
    ssh_channel_send_eof(channel);
    ssh_channel_close(channel);
    ssh_channel_free(channel);
    ssh_disconnect(session);
    ssh_free(session);
    return mReturnCode == 0;
}
} // namespace runner